import attr

import fp_globals
from .common import ConfigObject
from .instance import InstanceObject


@attr.s
class AccountObject(object):
    handle = attr.ib()
    auth = attr.ib()
    enabled = attr.ib(default=True)
    api_path = attr.ib(default=None)

    def __attrs_post_init__(self):
        self.userName, self.domain = self.handle.split("@")
        self.logged_in = False
        self.mpyInstance = None

        # TODO: have this be set by the account creator
        self.accountType = "mastodon\mastodonpy"

        #gui-related setup
        from fe_gui.widgets import accountSubmenu
        accmenu = fp_globals.Application.accountsMenu
        accmenu.addMenu(accountSubmenu(accmenu, self))
        self.notificationTabs = []

    async def log_in(self):
        from fe_core.external.mastodon import Mastodon
        #TODO: Move logic to mpy_wrapper and call based on self.accountType
        if self.api_path is None:
            api_base_url = self.domain
        else:
            api_base_url = self.api_path
        print(self.auth)
        self.mpyInstance = await Mastodon(api_base_url=api_base_url,
                                          client_id=self.auth[0],
                                          client_secret=self.auth[1],
                                          access_token=self.auth[3])
        self.mpyClient = self.mpyInstance  # fuck
        self.myUserDict = await self.mpyInstance.me()
        try:
            self.instance = fp_globals.loadedInstances[self.domain]
        except KeyError:
            # TODO: Properly instantiate instances with vanity domain handles when logging in
            fp_globals.loadedInstances[self.domain] = InstanceObject(self.domain)
            self.instance = fp_globals.loadedInstances[self.domain]
        finally:
            print("Logged into loaded instance %s" % self.instance.handle)
        self.logged_in = True

        from fe_core.mpy_wrapper import makempyNotificationNative
        nolist = await self.mpyClient.notifications()
        nolist.reverse()
        for no in nolist:
            nativeNotification = makempyNotificationNative(no, self.domain)
            for i in self.notificationTabs:
                i.addNotification(nativeNotification, self)
            fp_globals.unifiedNotificationWidget.addNotification(nativeNotification, self)

    async def getStatusIdWithSearchIfNeeded(self, status):
        try:
            return status.ids[self.domain]
        except KeyError:
            # Attempt to make instance aware of status
            mpyResult = await self.mpyInstance.search_v2(status.uri, result_type="statuses")

            if len(mpyResult["statuses"]):
                status.ids[self.domain] = str(mpyResult["statuses"][0]["id"])
                return status.ids[self.domain]
            else:
                # TODO: Mark a failed job
                fp_globals.systemMessage("Couldn't make %s aware of %s" % (self.domain, status.uri))
                raise

    async def getActorIdWithSearchIfNeeded(self, actor):
        try:
            return actor.ids[self.domain]
        except KeyError:
            # Attempt to make instance aware of status
            mpyResult = await self.mpyInstance.search_v2(actor.uri, result_type="accounts")

            if len(mpyResult["accounts"]):
                actor.ids[self.domain] = str(mpyResult["accounts"][0]["id"])
                return actor.ids[self.domain]
            else:
                # TODO: Mark a failed job
                fp_globals.systemMessage("Couldn't make %s aware of %s" % (self.domain, actor.uri))
                raise

    async def postStatus(self, status, in_reply_to_id=None, **kw):
        if self.accountType == "mastodon\mastodonpy":
            if in_reply_to_id:
                kw["in_reply_to_id"] = await self.getStatusIdWithSearchIfNeeded(in_reply_to_id)
            await self.mpyInstance.status_post(status, **kw)

    async def likeStatus(self, status):
        if self.accountType == "mastodon\mastodonpy":
            targetId = await self.getStatusIdWithSearchIfNeeded(status)
            await self.mpyInstance.status_favourite(targetId)

    async def boostStatus(self, status):
        if self.accountType == "mastodon\mastodonpy":
            targetId = await self.getStatusIdWithSearchIfNeeded(status)
            await self.mpyInstance.status_reblog(targetId)

    async def followActor(self, actor):
        if self.accountType == "mastodon\mastodonpy":
            targetId = await self.getActorIdWithSearchIfNeeded(actor)
            await self.mpyInstance.account_follow(targetId)

    def __str__(self):
        return self.handle

    def log_out(self):
        pass

    class maximumConcurrentWorkersPerAccountConfigObject(ConfigObject):
        name = "maximumConcurrentWorkersPerAccount"
        displayString = "Busy at most this many workers for requests to this account"
        selectionType = "int"
        default = 3

    class preselectInAccountSelectorConfigObject(ConfigObject):
        name ="preselectInAccountSelector"
        displayString = "Mark this account by default when posting, replying, liking, or boosting"
        selectionType = "bool"
        default = False



    configObjects = [maximumConcurrentWorkersPerAccountConfigObject,
                     preselectInAccountSelectorConfigObject]

    # Tried a dictionary comprehension. Ran into a serious scope problem.
    configObjectsByName = {}
    for x in configObjects:
        configObjectsByName[x.name] = x