import attr
from enum import Enum
from io import BytesIO
from PIL import Image

import fp_sqlite


@attr.s
class ConfigObject(object):
    context = attr.ib(default="_global_")
    configValue = attr.ib()

    @configValue.default
    def _defaultToDBOrDefault(self):
        try:
            return fp_sqlite.loadConfigValue(self.name, context=self.context)
        except LookupError:
            return self.default

    def __attrs_post_init__(self):
        # Whether option has changed *since it was loaded from DB*
        self.dirty = False

    @property
    def isSetToDefault(self):
        return self.configValue == self.default

    def resetToFactoryDefault(self):
        self.configValue = self.default
        #TODO: Also erase from DB

    def resetToUserDefault(self):
        try:
            return fp_sqlite.loadConfigValue(self.name, context=self.context)
        except LookupError:
            self.configValue = self.default
        #TODO: Also erase from DB

    def save(self):  # there's a joke here. I'll let you write it.
        fp_sqlite.syncConfigObjects(configObjects=self, force=True)

    def __repr__(self):
        return self.configValue


@attr.s
class MonitorObject(object):
    name = attr.ib()
    context = attr.ib()

    def __attrs_post_init__(self):
        self._dirty = False

    class monitorIsEnabledConfigObject(ConfigObject):
        name = "monitorIsEnabled"
        displayString = "i stg if you need that described further..."
        selectionType = "bool"
        default = False

    class monitorSaveStatusesConfigObject(ConfigObject):
        name = "monitorSaveStatuses"
        displayString = "Save statuses that are discovered with this monitor"
        selectionType = "bool"
        default = False

    class monitorCacheImagesConfigObject(ConfigObject):
        name = "monitorCacheImages"
        displayStriing = "Images from statuses are cached and thumbnailed"
        selectionType = "bool"
        default = False


@attr.s(slots=True)
class MetadataAttribute(object):
    v = attr.ib()
    fs = attr.ib()
    ls = attr.ib()
    dirty = attr.ib(default=True)

    @property
    def data(self):
        return self.v

    @data.setter
    def data(self, value, asOf=None):
        if asOf > self.ls:
            if value != self.v:
                self.dirty = True
                self.v = value
                self.fs = asOf

            self.ls = asOf

        #TODO: Mark dirty if lastseen is far later than value in db
        #Ensures DB lastseen value is kept up to date without excess writes


class cacheImageThumbnailsConfigObject(ConfigObject):
    name = "cacheImageThumbnails"
    displayString = "Make thumbnails for images and store them locally"
    selectionType = "bool"
    default = True


class statusesAreEphemeralConfigObject(ConfigObject):
    name = "statusesAreEphemeral"
    displayString = "Do not save statuses to database on disk."
    selectionType = "bool"
    default = False


class ignoreContentWarningsConfigObject(ConfigObject):
    name = "ignoreContentWarnings"
    displayString = "Always show content marked NSFW or with a content warning"
    selectionType = "bool"
    default = False


class VisibilityScope(Enum):
    PUBLIC = 0
    UNLISTED = 1
    PRIVATE = 2
    DIRECT = 3


#TODO: balete, it's been moved to new_workers
def thumbnailImage(imageData, imageFormat, size=150):
    imageObject = Image.open(BytesIO(imageData))
    imageObject.thumbnail((size, size))
    imageBlob = BytesIO()
    imageObject.save(imageBlob, format=imageFormat)
    return imageBlob