import asyncio
import json
import time
import httpx

import fp_globals
#OLD WORKERS MODULE


async def mpyWorker(mpyQueueOrder, accountHandle="anonymous"):
    while True:
        for queueKey in mpyQueueOrder:
            try:
                #does work from below line
                #jobToInsert = [mpyFunction, mpyArgs, mpyQueue, mpyCallback, target]
                mpyJob = fp_globals.mpyQueues[accountHandle][queueKey].get_nowait()
                try:
                    returnFromJob = await mpyJob[0](**mpyJob[1])
                except Exception as e:
                    print("mpy worker failed during callback")
                    print(mpyJob)
                    fp_globals.unifiedNotificationsTab.addSystemNotification(repr(e), "mpyWorker #69", "ERROR")
                    break
                
                fp_globals.mpyQueues[accountHandle][queueKey].task_done()
                mpyJob[3](returnFromJob)
                break #Do not wait, and also look for work on the high priority queues
            except asyncio.QueueEmpty:
                continue
        await asyncio.sleep(1)


async def httpWorker(httpQueues, httpQueueOrder):
    while True:
        for queueKey in httpQueueOrder:
            try:
                #print("http worker about to get, qsize is %i" % httpQueues[queueKey].qsize())
                httpJobCaller, httpJobUri = httpQueues[queueKey].get_nowait()
                
                if httpJobUri == "fake":
                    print("its fake")
                    break #see break note below
                print("I'd be doing a request on %s" % httpJobUri)

                async with httpx.ClientSession() as session:
                    async with session.get(httpJobUri) as response:
                        assert response.status == 200
                        jsonFromRequest = await response.read()
                        translatedTimeline = []
                        for i in json.loads(jsonFromRequest):
                            translatedTimeline.append(statusObjectFromMastodonDict(i))
                            
                httpQueues[queueKey].task_done()
                httpJobCaller.httpWorkerCallback(translatedTimeline)
                break #Do not wait, and also look for work on the high priority queues
            except asyncio.QueueEmpty:
                continue #Go to next queue

        #Queues exhausted.
        await asyncio.sleep(10)


async def mastodonStreamingMonitorWorker(assignedInstances):
    from mpy_wrapper import mpyMonitorStreamCallback
    from external.mastodon import Mastodon
    from random import randint
    loop = asyncio.get_event_loop()
    await asyncio.sleep(randint(2,8)*0.1)
    while True:
        if fp_globals.Application.firehoseAction.isChecked():
            #Look for and assign self at most one "new" instance per cycle
            for i in fp_globals.loadedInstances.values():
                if i._monitorhttpxClient == None:
                    if i.config["includeInFirehose"]:
                        assignedInstances[i] = None
                        i._monitorhttpxClient = "LOCK"
                        break

            copyOfInstances = assignedInstances.keys()

            for i in copyOfInstances:
                try:
                    if assignedInstances[i] == None:
                        if i._monitorhttpxClient == None or "LOCK":
                            timeSinceLastStreamingError = time.time() - i.lastStreamingError
                            if timeSinceLastStreamingError > i.config["mastodonStreamingSecondsToBackoff"].configValue:
                                i._monitorhttpxClient = i.mkhttpxClient(monitor=True)
                            else:
                                continue

                        mpyclient = await Mastodon(api_base_url=i.domain, session=i._monitorhttpxClient)
                        assignedInstances[i] = asyncio.create_task(mpyclient.stream_public(mpyMonitorStreamCallback(str(i.domain), None),
                                                                                           run_async=True,
                                                                                           reconnect_async=False))

                        continue
                    elif type(assignedInstances[i]) == asyncio.Task:
                        pass
                        #TODO: check health of connection
                        #print("monitor httpx instance already exists and should be running")
                    else:
                        print("monitor httpx instance isn't an asyncio.task it's a %s" % str(type(assignedInstances[i])))
                except Exception as e:
                    #fp_globals.systemMessage("Failed to set up monitor for %s because %s" % (i.handle, repr(e)))
                    print("About to try reporting streaming failure")
                    i.streamingFailed(e)
        await asyncio.sleep(2)

def schedulempyJob(mpyFunction, mpyArgs, mpyQueue, mpyCallback, target="anonymous", do_in=0):
    if do_in == 0: #Run job NOW, don't schedule.
        jobToInsert = [mpyFunction, mpyArgs, mpyQueue, mpyCallback, target]
        fp_globals.mpyQueues[target][mpyQueue].put_nowait(jobToInsert)

    else:
        jobToInsert = [int(time.time())+do_in, mpyFunction, mpyArgs, mpyQueue, mpyCallback, target]
        
        try:
            fp_globals.mpySchedules[target].append(jobToInsert)
        except Exception as e:
            print(e)
        
async def mpyJobSchedulerLoop():
    while True:
        for target in fp_globals.mpySchedules.keys():
            #.copy() to make a new list, so we can delete from the original
            for i in fp_globals.mpySchedules[target].copy():
                if i[0] < int(time.time()):
                    #Strip timing data, insert object
                    print("queueing a thing.")
                    fp_globals.mpyQueues[target][i[3]].put_nowait(i[1:])
                    fp_globals.mpySchedules[target].remove(i)
            #print("Scheduled %s tasks: %i" % (target,len(fp_globals.mpySchedules[target])))
        
        await asyncio.sleep(.2)
        
def kickoffAccountWorker(account):
    accountHandle = account.handle
    fp_globals.mpyQueues[accountHandle] = {'timelineFetchQueue': asyncio.Queue(),
                                           'statusCrawlQueue': asyncio.Queue()}
    
    mpyWorkers[accountHandle] = [asyncio.get_event_loop().create_task(mpyWorker(mpyQueueOrder, accountHandle=accountHandle))]
    print(mpyWorkers)
    fp_globals.mpySchedules[accountHandle] = []


def dispatchStreamingMonitor(instance):
    from random import choice
    try:
        mpyFirehoseAssignments = fp_globals.mpyFirehoseAssignments

        choice(list(mpyFirehoseAssignments.values()))[instance.handle] = None
    except Exception as e:
        print("!!!worker problem adding monitor %s" % repr(e))
    
        
def initWorkers():
    global mpyQueueOrder
    global mpyWorkers
    ##Queues##

    #Workers will perform work in the following order.
    #The first-listed queues must be empty before work is done from others
    httpQueueOrder = ['timelineFetchQueue']
    mpyQueueOrder = ['statusCrawlQueue',  'timelineFetchQueue', 'backgroundStatusCrawlQueue']
    
    #Start http workers
    httpQueues = {'timelineFetchQueue': asyncio.Queue()}
    httpWorkers = [asyncio.create_task(httpWorker(httpQueues, httpQueueOrder)) for _ in range(2)]
    
    #Start anonymous mpy workers
    mpyWorkers = {}
    
    fp_globals.mpyQueues["anonymous"] = {'timelineFetchQueue': asyncio.Queue(),
                                         'statusCrawlQueue': asyncio.Queue(),
                                         'backgroundStatusCrawlQueue': asyncio.Queue()}
    mpyWorkers["anonymous"] = [asyncio.create_task(mpyWorker(mpyQueueOrder)) for _ in range(10)]
    mpyJobSchedulerWorker = asyncio.create_task(mpyJobSchedulerLoop())
    
    #Start logged-in workers
    print("Starting logged-in workers")
    for i in fp_globals.loadedAccounts.values():
        print("h")
        kickoffAccountWorker(i)
    #    fp_globals.mpyQueues[i.handle] = {'timelineFetchQueue': asyncio.Queue()}
    #    mpyWorkers[i.handle] = [loop.create_task(mpyWorker(mpyQueueOrder, accountHandle=i.handle))]
    #    fp_globals.mpySchedules[i.handle] = []

    #using main event loop for firehose for now
    #it really should be in a separate process
    fp_globals.mpyFirehoseAssignments = {}
    mpyFirehoseAssignments = fp_globals.mpyFirehoseAssignments
    for _ in range(2):
        newAssignmentDict = {}
        mpyFirehoseAssignments[asyncio.create_task(mastodonStreamingMonitorWorker(newAssignmentDict))] = newAssignmentDict