#!/usr/bin/python
import asyncio
from PyQt5.QtWidgets import QMainWindow, QTabWidget, QAction, QTabBar, QLabel
from qasync import asyncClose
from contextlib import suppress
from fe_gui import icons, subwindows
from fe_gui.actions import initActions
from fe_gui.tabs.__init__ import *
from fe_gui.tabs.mastodon_tabs import *
from fe_core import fp_sqlite, fp_globals
from fe_core.fe_types import ConfigObject


class App(QMainWindow):
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        self.setMinimumSize(1200, 800)
        self.setWindowTitle(fp_globals.productName)
        self.statusBarLabelWidget = QLabel("Initializing...")
        self.statusBar().addWidget(self.statusBarLabelWidget)

        initActions(self)

        self.mainToolBar = self.addToolBar("")
        self.mainToolBar.addAction(self.openNewStatusDialogAction)
        self.mainToolBar.addAction(self.addTabAction)
        self.mainToolBar.addAction(self.searchAction)
        self.mainToolBar.addAction(self.settingsAction)
        self.mainToolBar.addAction(self.firehoseAction)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addAction(self.exitAction)

        accountsMenu = menubar.addMenu('&Accounts')
        self.accountsMenu = accountsMenu
        accountsMenu.addAction(self.addAccountAction)

        debugMenu = menubar.addMenu('&Debug')
        debugMenu.addAction(self.syncDBAction)
        debugMenu.addAction(self.saveTabsAction)
        debugMenu.addAction(self.marryAmericanWomanAction)

        self.tabs = QTabWidget()
        self.tabs.setTabsClosable(True)
        self.tabs.tabCloseRequested.connect(self.tabCloseRequested)
        self.homeTab = newHomeTab(self)
        self.systemNotificationsTab = notificationTab(self)
        fp_globals.systemNotificationWidget = self.systemNotificationsTab.notificationWidget
        fp_globals.unifiedNotificationWidget = self.homeTab.notificationWidget

        [self.tabs.addTab(x, "...") for x in (self.homeTab, self.systemNotificationsTab)]
        for i,w in enumerate([self.tabs.widget(i) for i in range(self.tabs.count())]):
            w.updateTabName()
            if w.systemTab:
                self.tabs.tabBar().setTabButton(i, QTabBar.RightSide, None)

        self.setCentralWidget(self.tabs)
        self.showMaximized()

    def tabCloseRequested(self, index):
        if not self.tabs.widget(index).systemTab:
            self.tabs.removeTab(index)

    @asyncClose
    async def closeEvent(self, event):
        fp_globals.shuttingDown = True
        self.shuttingDown = subwindows.shuttingDownDialog(self)

        waitfor = []
        for i in fp_globals.loadedInstances.values():
            if i._httpxClient is not None:
                waitfor.append(asyncio.create_task(i._httpxClient.aclose()))
            if i._monitorhttpxClient is not None:
                waitfor.append(asyncio.create_task(i._monitorhttpxClient.aclose()))
        # I hate this, too. Not as much as I hate me, though!
        await asyncio.sleep(2)
        while not all([f.done() for f in waitfor]):
            with suppress(asyncio.CancelledError):
                for f in waitfor:
                    f.cancel()
                    await f
            await asyncio.sleep(.2)
        if fp_globals.Application.config["saveDatabaseOnExit"].configValue:
            fp_sqlite.syncDirtyObjects()
        if fp_globals.Application.config["savePersistentTabsOnExit"].configValue:
            fp_sqlite.savePersistentTabs()

    def firehoseActionChanged(self):
        if self.firehoseAction.isChecked():
            self.firehoseTab = firehoseTimelineTab(self, context=None)
            self.tabs.addTab(self.firehoseTab, "...")
            self.firehoseTab.updateTabName()
            for instance in fp_globals.loadedInstances.values():
                print("Considering %s" % instance.handle)
                if instance.config["includeInFirehose"].configValue:
                    if instance.config["disableStreaming"].configValue:
                        break  # A different type of worker will handle this type of monitor

                    if instance._monitorhttpxClient == None:
                        print("Trying to add monitor for %s" % instance.handle)
                        workers.dispatchStreamingMonitor(instance)

            # Didn't write the logic to stop it
            # this train has no brakes --r000t 25dec2020

    def showNewVersionAvailableDialog(self, latestVersion):
        w = subwindows.newVersionAvailableDialog(self, latestVersion)

    def savePersistentTabs(self):
        fp_sqlite.savePersistentTabs(self.tabs)

    def marryAmericanWoman(self):
        w = subwindows.marryAmericanWomanDialog(self)



    def addNewTabWithDialog(self):
        w = subwindows.addTabDialog(self)

    def openSearchDialog(self):
        w = subwindows.addSearchTabDialog(self)

    def openGlobalSettingsDialog(self):
        # TODO: Make phrasing consistent
        w = subwindows.globalSettingsWindow(self)

    def updateStatusBar(self):
        self.statusBarLabelWidget.setText("%s 0.0.%i - %i statuses and %i actors in memory" % (fp_globals.productName,
                                                                                               fp_globals.productVersion,
                                                                                               len(
                                                                                                   fp_globals.loadedStatuses),
                                                                                               len(
                                                                                                   fp_globals.loadedActors)))


async def guiStartup(h):
    icons.load()
    h.statusBarLabelWidget.setText("Loading persistent tabs...")
    for count, tabToLoad in enumerate(fp_sqlite.loadPersistentTabs(), start=0):
        context = tabToLoad[2]
        tabClass = globals()[tabToLoad[1]]
        configContext = "_tab_%i_" % count
        h.tabs.addTab(tabClass(h.tabs, context), "Loading persistent tab...")
    for i in [h.tabs.widget(i) for i in range(h.tabs.count())]:
        i.updateTabName()


class savePersistentTabsOnExitConfigObject(ConfigObject):
    name = "savePersistentTabs"
    displayString = "Sync persistent tabs when exiting"
    selectionType = "bool"
    default = True