#!/usr/bin/python
from PyQt5.QtWidgets import QPushButton, QTableWidgetItem, \
    QTableWidget, QLabel, QWidget, QVBoxLayout, \
    QHBoxLayout, QAbstractItemView, QHeaderView, QLineEdit, QFrame, QGroupBox, QListWidget
from collections import deque
from random import randrange
from time import time
from datetime import datetime
import fe_core.fe_types.common
from fe_gui import widgets
from fe_core import fp_globals, workers


class newHomeTab(QWidget):
    #fearing the canceldon admins no longer, we moved out of our old default tab, "HOMETAB"
    #we braved armchair activists, smug neo-karens, gargron, and lukewarm takes, until we
    #reached what now shows up when you launch fediEngine_gui for the first time, "NEWHOMETAB"
    systemTab = True
    def __init__(self, parent, **kw):
        super(QWidget, self).__init__(parent)
        self.tabIsPersistent = False

        self.emptyServerBanner = QFrame()
        self.emptyServerBanner.setObjectName("emptyServerBanner")
        self.emptyServerBanner.setStyleSheet("#emptyServerBanner {background-color:blue;}")
        self.emptyServerBannerLayout = QHBoxLayout()
        self.emptyServerBanner.setLayout(self.emptyServerBannerLayout)
        self.emptyServerBannerText = QLabel(
            'fediEngine is not preloaded with any instances. You must supply at least one instance, or a list.')
        self.emptyServerBannerLayout.addWidget(self.emptyServerBannerText)

        self.notificationWidget = widgets.notificationWidget()
        self.notificationLayout = QVBoxLayout()
        self.notificationLayout.addWidget(self.notificationWidget)
        self.notificationGroupBox = QGroupBox("Notifications")
        self.notificationGroupBox.setLayout(self.notificationLayout)

        self.instanceListWidget = widgets.instanceListWidget(self)
        self.discoverInstanceButton = QPushButton("Discover instance...")
        self.discoverInstanceButton.clicked.connect(parent.openDiscoverInstanceDialog.triggered)
        self.importListButton = QPushButton("Import list...")
        self.refreshButton = QPushButton("Refresh")
        self.refreshButton.clicked.connect(self.instanceListWidget.refresh)
        self.serverListButtonLayout = QHBoxLayout()
        self.serverListButtonLayout.addWidget(self.discoverInstanceButton)
        self.serverListButtonLayout.addWidget(self.importListButton)
        self.serverListButtonLayout.addWidget(self.refreshButton)
        self.serverListLayout = QVBoxLayout()
        self.serverListLayout.addLayout(self.serverListButtonLayout)
        self.serverListLayout.addWidget(self.instanceListWidget)
        self.serverListGroupBox = QGroupBox("Instance Browser")
        self.serverListGroupBox.setLayout(self.serverListLayout)

        self.directMessagesWidget = QListWidget()
        self.directMessagesLayout = QVBoxLayout()
        self.directMessagesLayout.addWidget(self.directMessagesWidget)
        self.directMessagesGroupBox = QGroupBox("Direct Messages")
        self.directMessagesGroupBox.setLayout(self.directMessagesLayout)

        self.newHomeLayout = QHBoxLayout()
        self.newHomeLayout.addWidget(self.notificationGroupBox)
        self.newHomeLayout.addWidget(self.directMessagesGroupBox)
        self.newHomeLayout.addWidget(self.serverListGroupBox)
        self.newHomeLayout.setStretch(0,5)
        self.newHomeLayout.setStretch(1,3)
        self.newHomeLayout.setStretch(2,2)

        self.layout = QVBoxLayout()
        self.layout.addLayout(self.emptyServerBannerLayout)
        self.layout.addLayout(self.newHomeLayout)
        self.setLayout(self.layout)

    def _tabName(self):
        return "Home"

    def updateTabName(self):
        tabHolder = self.parent().parent()
        thisTabIndex = tabHolder.indexOf(self)
        print("index %i" % thisTabIndex)
        tabHolder.setTabText(thisTabIndex, self._tabName())


class notificationTab(QWidget):
    systemTab = True

    def __init__(self, parent, **kw):
        super(QWidget, self).__init__(parent)
        self.tabIsPersistent = False
        self.notificationWidget = widgets.notificationWidget()
        dataLayout = QVBoxLayout()
        dataLayout.addWidget(self.notificationWidget)
        self.setLayout(dataLayout)

    def _tabName(self):
        return "System Log"
        
    def updateTabName(self):
        tabHolder = self.parent().parent()
        thisTabIndex = tabHolder.indexOf(self)
        print("index %i" % thisTabIndex)
        tabHolder.setTabText(thisTabIndex, self._tabName())

class timelineTab(QWidget):
    showFetchButtons = True
    systemTab = False
    def __init__(self, parent, context, config=None, configContext=None, **kwargs):
        super(QWidget, self).__init__(parent)
        self.context = context
        if configContext is None:
            self.configContext = "_tab_%016x_" % randrange(16**16)
        else:
            self.configContext = configContext
        self.initializeConfig(config)
        self.nextPagesLeft = 0
        self.timeline = deque()
        self.dataView = widgets.statusListDataView()

        if self.showFetchButtons:
            self.nextButtonsLayout = QHBoxLayout()
            self.nextPageButton = QPushButton("Fetch Next Page", self)
            self.next5PagesButton = QPushButton("Fetch Next 5 Pages", self)
            self.next20PagesButton = QPushButton("Fetch Next 20 Pages, you heathen.", self)
            self.nextPageButton.clicked.connect(self.fetchNextPageButtonPressed)
            self.next5PagesButton.clicked.connect(self.fetchNext5PagesButtonPressed)
            self.next20PagesButton.clicked.connect(self.fetchNext20PagesButtonPressed)
            self.nextButtonsLayout.addWidget(self.nextPageButton)
            self.nextButtonsLayout.addWidget(self.next5PagesButton)
            self.nextButtonsLayout.addWidget(self.next20PagesButton)

        #old tab toolbar that was obviously never gonna be a toolbar
        #but also toolbars can't be attached to things that aren't the main window
        #so maybe this has to stay until it's all in the same main toolbar --r000t 18dec2020
        self.tabStatusBoxLayout = QHBoxLayout()
        self.refreshButton = QPushButton("Refresh (Waiting to run)", self)
        self.hideBoostsButton = QPushButton("Hide boosts", self)
        self.hideBoostsButton.setCheckable(True)
        self.quickFilterTextBox = QLineEdit()
        self.quickFilterTextBox.setPlaceholderText("Quick Filter")
        self.quickFilterTextBox.setClearButtonEnabled(True)
        self.quickFilterTextBox.textChanged.connect(self.quickFilterTextChanged)
        self.tabStatusBoxLayout.addWidget(self.refreshButton)
        self.tabStatusBoxLayout.addWidget(self.hideBoostsButton)
        self.tabStatusBoxLayout.addWidget(self.quickFilterTextBox)

        #Maybe subclass it? 2020jul15
        self.twoPaneTabLayout = QHBoxLayout()
        self.statusDetailsSidebar = QVBoxLayout()
        self.statusDetailsPlaceholderLabel = QLabel("Details of the selected status appear here")
        self.statusDetailsSidebar.addWidget(self.statusDetailsPlaceholderLabel)
        self.dataView.currentItemChanged.connect(self.selectedStatusChanged)
        #Put view in tab
        self.dataLayout = QVBoxLayout()
        self.errorBanner = QFrame()
        self.errorBanner.setObjectName("errorBanner")
        self.errorBanner.setStyleSheet("#errorBanner {background-color:red;}")
        self.errorBannerLayout = QHBoxLayout()
        self.errorBanner.setLayout(self.errorBannerLayout)
        self.errorBanner.setVisible(False)
        self.errorBannerText = QLabel()
        self.errorBannerLayout.addWidget(self.errorBannerText)
        self.dataLayout.addWidget(self.errorBanner)
        self.dataLayout.addLayout(self.tabStatusBoxLayout)
        self.dataLayout.addWidget(self.dataView)
        if self.showFetchButtons:
            self.dataLayout.addLayout(self.nextButtonsLayout)

        self.twoPaneTabLayout.addLayout(self.dataLayout, stretch=3)
        self.twoPaneTabLayout.addLayout(self.statusDetailsSidebar)
        self.setLayout(self.twoPaneTabLayout)

    def initializeConfig(self, configFromCaller):
        self.config = {}
        for i in self.configObjects:
            try:
                self.config[i.name] = i(configValue=configFromCaller[i.name], context=self.configContext)
            except (KeyError, TypeError):
                self.config[i.name] = i(context=self.configContext)
        
    def _tabName(self):
        return "..."

    def showErrorBanner(self, message):
        self.errorBannerText.setText(str(message))
        self.errorBanner.setVisible(True)

    def updateTabName(self):
        thisTabIndex = fp_globals.Application.tabs.indexOf(self)
        fp_globals.Application.tabs.setTabText(thisTabIndex, self._tabName())

    def updateRefreshButton(self, text, ts=time()):
        stringTimestamp = datetime.fromtimestamp(ts).strftime("%I:%M:%S %p")
        self.refreshButton.setText("Refresh (%s)" % text % stringTimestamp)

    def addStatusToFront(self, status, **kw):
        return self.addStatus(status, appendToBack=True, **kw)

    def addStatus(self, status, fetchContext=True, appendToBack=False):
        #self.newestStatus = status
        if appendToBack:
            self.timeline.append(status)
        else:
            self.timeline.appendleft(status)
        self.dataView.addStatus(status)
        if fetchContext:
            workers.schedulempyJob(status.fetchContext, {"deep":False}, "backgroundStatusCrawlQueue", fp_globals.doFuckAll)

    def quickFilterTextChanged(self):
        self.dataView.updateModel()

    def fetchNextPageButtonPressed(self, numberOfPages=1):
        self.fetchNextPages(1)
    
    def fetchNext5PagesButtonPressed(self):
        self.fetchNextPages(5)
    
    def fetchNext20PagesButtonPressed(self):
        self.fetchNextPages(20)
        
    def fetchNextPages(self, **kw):
        #This function is meant to be overridden by subclasses
        pass

    def refreshButtonPressed(self, **kw):
        #This function is meant to be overridden by subclasses
        pass

    def selectedStatusChanged(self, current, previous, **kw):
        from fe_gui.subwindows import statusDetailsDialog
        oldWidget = self.statusDetailsSidebar.itemAt(0).widget()
        newLayout = QVBoxLayout()
        newDetailsWidget = statusDetailsDialog(self, current.status, inline=True)
        newLayout.addWidget(newDetailsWidget)

        newWidget = QWidget()
        self.statusDetailsSidebar.replaceWidget(oldWidget, newWidget)
        newWidget.setLayout(newLayout)
        oldWidget.setParent(None)
        oldWidget = None

    class tabIsPersistentConfigObject(fe_core.fe_types.common.ConfigObject):
        name = "tabIsPersistent"
        displayString = "Keep this tab around the next time %s starts" % fp_globals.productName
        selectionType = "bool"
        default = True

    class refreshIntervalConfigObject(fe_core.fe_types.common.ConfigObject):
        name = "refreshInterval"
        displayString = "Seconds between polls, when polling"
        selectionType = "int"
        default = 15

    class hideBoostsConfigObject(fe_core.fe_types.common.ConfigObject):
        name = "hideBoosts"
        displayString = "Do not show statuses that were boosted into this timeline"
        selectionType = "bool"
        default = False

    class disableStreamingConfigObject(fe_core.fe_types.common.ConfigObject):
        name = "disableStreaming"
        displayString = "Do not use the Mastodon streaming API"
        selectionType = "bool"
        default = False

    configObjects = [tabIsPersistentConfigObject,
                     refreshIntervalConfigObject,
                     hideBoostsConfigObject,
                     disableStreamingConfigObject]


class firehoseTimelineTab(timelineTab):
    systemTab = True
    showFetchButtons = False
    def _tabName(self):
        return "(local) Firehose"

    def addStatus(self, status, fetchContext=False):
        for i in reversed(range(self.dataView.invisibleRootItem().childCount())):
            widgetItem = self.dataView.invisibleRootItem().child(i)
            if widgetItem.status == status:
                return
        timelineTab.addStatus(self, status, fetchContext)


class searchResultTab(timelineTab):
    def __init__(self, parent, search, **kwargs):
        timelineTab.__init__(self, parent, search, **kwargs)
        self.search = search
        self.firstBuild = True
        
    def _tabName(self):
        return '(local) Search: "%s"' % self.search
        
    def updateModel(self):
        from fe_core.fp_sqlite import globalContentSearch
        if self.firstBuild:
            #Do database search only the first timeline
            #Any new statuses will be in RAM. 
            globalContentSearch(self.search)
            #DB should load relevant statuses into memory
            #now we can find them
            #but they are also returned by that function
    
        #TODO: make this a nice sexy list comprehension
        for uri in fp_globals.loadedStatuses:
            status = fp_globals.loadedStatuses[uri]
            if status.content.lower().find(self.search.lower()) != -1:
                self.addStatus(status)
        self.firstBuild = False

    def showEvent(self, h):
        self.updateModel()

