#!/usr/bin/python
from asyncio import get_event_loop
from fe_core.fe_types import *
from mpy_wrapper import *
from .__init__ import *


class mastodonTimelineTab(timelineTab):
    oneshot = False

    def __init__(self, parent, context, config=None, account=None, **kw):
        timelineTab.__init__(self, parent, context, config, **kw)
        #print("Starting mastodon tab with context %s of type %s" % (str(self.context), type(self.context)))
        self.account = account
        if type(self.context) is str:
            #TODO: See why this sometimes returns None
            self.context = self._str2realObj(self.context)
        if type(self.context) in (InstanceObject, ActorObject, StatusObject, AccountObject):
            self.domain = self.context.domain
        elif self.context is None:
            if type(context) is str:
                self.domain = context
                self.context = context
            else:
                print("Bailing lmao, context is of type" % type(context))
                return

        if not self.config["disableStreaming"].configValue:
            if type(self.context) is not ActorObject:
                self.startStreaming()
                self.oneshot = True
        self.startPollingJob()

    def mpyCallback(self, callbackTuple):
        nativeStatusList, mpyClient = callbackTuple
        try:
            if len(nativeStatusList) > 0:
                for i in nativeStatusList:
                    self.addStatus(i)
                self.dataView.updateModel()
            else:
                print("No new statuses.")
                self.nextPagesLeft = 0
        except Exception as e:
            fp_globals.systemMessage(repr(e), sender="mpyCallback", severity="WARN")
            return

        argDict = {'timeline': self.mpyScope,
                   'context': self.context,
                   'api_base_url': self.domain,
                   'mpyClient': mpyClient,
                   'account': self.account}

        if self.nextPagesLeft > 0:
            argDict["max_id"] = self.timeline[-1].ids[self.domain]
            self.nextPagesLeft -= 1
            print("%i pages after next job" % self.nextPagesLeft)
            workers.schedulempyJob(mastodonTimeline, argDict, "timelineFetchQueue", self.mpyCallback, do_in=0)
            return

        if not self.oneshot:
            #Schedule next refresh
            workers.schedulempyJob(mastodonTimeline, argDict, "timelineFetchQueue", self.mpyCallback, do_in=self.refetchInterval)
            return

    def _str2realObj(self, context):
        #This is meant to be overridden by subclasses
        #But it will try to load a domain if it's not
        try:
            return getInstanceObjectByHandle(context)

        except LookupError:
            print("Failed to get instance object for %s from DB" % context)
            return context

    def monitor(self, cause, **kwargs):
        if cause == "streamingAborted":
            self.updateRefreshButton("Streaming API died at %s")

    def startPollingJob(self):
        #TODO: make it safe to assume self.context is not a string anymore
        target = "anonymous"
        argDict = {'timeline': self.mpyScope,
                   'context': self.context}
        if type(self.context) == AccountObject:
            argDict["account"] = self.context
            argDict["domain"] = self.context.instance.domain
            target = self.context.handle
        elif type(self.context) == str:
            argDict["domain"] = self.context
        #workers.schedulempyJob(mastodonTimeline, argDict, "timelineFetchQueue", self.mpyCallback, target=target)
        pollingTask = get_event_loop().create_task(mastodonTimeline(**argDict))
        #pollingTask.add_done_callback(self.mpyCallback)
        self.updateRefreshButton("Worker started at %s")

    def startStreaming(self):
        loop = get_event_loop()
        if type(self.context) is AccountObject:
            mpyscope = "user"
        else:
            mpyscope = self.mpyScope
        loop.create_task(startStreamingThread(self.context,
                                              mpyscope,
                                              self.addStatusToFront,
                                              heartbeatReceiver=self.streamingHeartbeatReceiver))

    def streamingHeartbeatReceiver(self, ts):
        self.updateRefreshButton("Stream last active %s", ts=ts)


    def refreshButtonPressed(self, **kw):
        #TODO: Restart streaming if down
        self.startPollingJob()

    def fetchNextPages(self, numberOfPages):
        self.nextPagesLeft = numberOfPages

        argDict = {'timeline': self.mpyScope,
                   'api_base_url': self.domain,
                   'context': self.context}

        #TODO: See if catching IndexError instead is faster
        if self.timeline:
            try:
                lastStatus = self.timeline[-1]
                lastId = lastStatus.ids[self.domain]
            except KeyError:
                if type(self.context) is str:
                    lastId = lastStatus.ids[self.context]
                elif type(self.context) is InstanceObject:
                    lastId = lastStatus.ids[self.context.domain]
                else:
                    fp_globals.systemMessage("fetch failed, context was unknown type %s" % type(self.context))
                    return
            argDict['max_id'] = lastId

        try:
            argDict["account"] = self.account
        except AttributeError:
            pass
        print("nextpage about to schedule job")
        workers.schedulempyJob(mastodonTimeline, argDict, "timelineFetchQueue", self.mpyCallback)


class mastodonFederatedTimelineTab(mastodonTimelineTab):
    scopeString = "Public"
    mpyScope = "public"
    def _tabName(self):
        return "(anon) %s Public" % self.domain


class mastodonHomeTimelineTab(mastodonTimelineTab):
    mpyScope = "home"
    def _tabName(self):
        return "(%s) Home" % self.context.handle
    def _str2realObj(self, context):
        print("About to lookup actor string")
        return fp_globals.loadedAccounts[context]


class mastodonActorTab(mastodonTimelineTab):
    scopeString = "Local"
    mpyScope = None
    def _tabName(self):
        return ("(anon) %s" % self.context)


class mastodonLocalTimelineTab(mastodonTimelineTab):
    scopeString = "Local"
    mpyScope = "local"
    def _tabName(self):
        return "(anon) %s Local" % self.context