from asyncio import create_task, get_event_loop
from datetime import datetime
from io import BytesIO

from PyQt5.QtWidgets import QWidget, QLabel, QHBoxLayout, QVBoxLayout, QGridLayout, QPlainTextEdit, QTreeWidget, \
    QTreeWidgetItem, QTreeWidgetItemIterator, QMenu, QCheckBox, QGroupBox, QPushButton, QTabWidget, QListView, \
    QComboBox, QSpinBox, QDialogButtonBox, QLineEdit, QListWidget, QTableWidget, QHeaderView, QAbstractItemView, \
    QTableWidgetItem, QListWidgetItem, QLayout
from PyQt5.QtCore import (Qt, QObject)
from PyQt5.QtGui import QPixmap, QStandardItemModel, QStandardItem
from webbrowser import open as openBrowserTab

import fe_core.fe_types.account
import fe_core.fe_types.actor
import fe_core.fe_types.common
import fe_core.fe_types.instance
import fe_core.fe_types.status
from fe_core import fp_globals, fp_sqlite, workers
import attr

class notificationWidget(QTableWidget):
    TIMESTAMP,ACCOUNT,USER,ACTION,CONTENT = range(5)
    def __init__(self, **kw):
        QTableWidget.__init__(self, 0, 5)

        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.verticalHeader().setVisible(False)
        self.setAlternatingRowColors(True)

        self.setHorizontalHeaderLabels(["Timestamp",
                                        "Your Account",
                                        "Their Account",
                                        "Action",
                                        "Content"])
        header = self.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(3, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(4, QHeaderView.Stretch)

    def addSystemNotification(self, message, sender="Log Message", severity="NOTICE"):
        self.insertRow(0)
        self.setItem(0, self.TIMESTAMP, QTableWidgetItem(str(datetime.now()).split('.')[0]))
        self.setItem(0, self.ACCOUNT, QTableWidgetItem("fediEngine@localhost"))
        self.setItem(0, self.USER, QTableWidgetItem(str(sender)))
        self.setItem(0, self.ACTION, QTableWidgetItem(str(severity)))
        self.setItem(0, self.CONTENT, QTableWidgetItem(str(message)))

    def addNotification(self, notification, account):
        self.insertRow(0)
        self.setItem(0, self.TIMESTAMP, QTableWidgetItem(str(datetime.now()).split('.')[0]))
        self.setItem(0, self.ACCOUNT, QTableWidgetItem(account.handle))

        self.setItem(0, self.USER, QTableWidgetItem(str(notification["source"])))
        self.setItem(0, self.ACTION, QTableWidgetItem(notification["type"]))
        if notification["type"] != "follow":
            self.setItem(0, self.CONTENT, QTableWidgetItem(notification["status"].sanitizedContent))


class instanceListWidget(QListWidget):
    def __init__(self, parent):
        QListWidget.__init__(self, parent)

        self.initModel()

    def initModel(self):
        mt = fe_core.InstanceMetadataType
        for i in fp_globals.loadedInstances.values():
            try:
                upperLabel = QLabel("%s - %s" % (i.handle, i.meta[mt.TITLE]))
            except AttributeError:
                upperLabel = QLabel(i.handle)
            try:
                lowerLabel = QLabel("%s users, %i peers" % (i.meta[mt.ADVERTISED_STATUS_COUNT],
                                                            i.meta[mt.ADVERTISED_PEER_COUNT]))
            except AttributeError:
                lowerLabel = QLabel("Awaiting discovery...")

            iwl = QVBoxLayout()
            iwl.addWidget(upperLabel)
            iwl.addWidget(lowerLabel)
            iwl.addStretch()
            iwl.setSizeConstraint(QLayout.SetFixedSize)
            iww = QWidget()
            iww.setLayout(iwl)
            ili = QListWidgetItem()
            ili.setSizeHint(iww.sizeHint())
            self.addItem(ili)
            self.setItemWidget(ili, iww)

    def refresh(self):
        #TODO: Attach widgets to instance before clearing, so they don't need to be rebuilt
        #Or maybe do something other than rebuild the whole fucking thing.
        self.clear()
        self.initModel()


class accountSubmenu(QMenu):
    def __init__(self, parent, account):
        QMenu.__init__(self, account.handle, parent)
        self.account = account
        self.addAction("New Status...")

    def newStatusFromAccount(self):
        from fe_gui.subwindows import newStatusDialog
        h = newStatusDialog(fp_globals.Application, accounts=(self.account,))


class cachedImageThumbnailWidget(QLabel):
    def __init__(self, parent, image):
        QLabel.__init__(self, parent)
        self.image = image
        self.mousePressEvent = self.openInBrowser

    async def load(self):
        blob = await get_event_loop().run_in_executor(fp_globals.Application.ioPool,
                                                      fp_sqlite.fetchCachedImageIfExists,
                                                      self.image.uri)
        self.iPixmap = QPixmap()
        self.iPixmap.loadFromData(blob)
        self.setPixmap(self.iPixmap)

    def openInBrowser(self, *a, **kw):
        openBrowserTab(self.image.uri)


class actorTasksGroupBox(QGroupBox):
    def __init__(self, parent):
        QGroupBox.__init__(self, "Actor Tasks")
        self.actorActionsLayout = QVBoxLayout()

        self.showActorWindowAction = QPushButton("View Actor...")
        self.showActorWindowAction.clicked.connect(parent.openActorDetailsDialog)
        self.followActorAction = QPushButton("Follow Actor from Accounts...")
        self.followActorAction.clicked.connect(parent.openFollowActorDialog)
        self.openActorTabAction = QPushButton("Open Actor in Anonymous Tab")
        self.openActorTabAction.clicked.connect(parent.openActorTab)
        self.muteActorAction = QPushButton("Mute Actor...")
        self.blockActorAction = QPushButton("Block Actor...")

        self.actorActionsLayout.addWidget(self.showActorWindowAction)
        self.actorActionsLayout.addWidget(self.followActorAction)
        self.actorActionsLayout.addWidget(self.openActorTabAction)
        self.actorActionsLayout.addWidget(self.muteActorAction)
        self.actorActionsLayout.addWidget(self.blockActorAction)
        self.setLayout(self.actorActionsLayout)


class statusTasksGroupBox(QGroupBox):
    def __init__(self, parent):
        QGroupBox.__init__(self, "Status Tasks")
        self.statusActionsLayout = QVBoxLayout()

        self.replyStatusAction = QPushButton("Reply from Accounts...")
        self.replyStatusAction.clicked.connect(parent.openReplyToStatusDialog)
        self.quoteReplyStatusAction = QPushButton("Quote Reply from Accounts...")
        self.quoteReplyStatusAction.clicked.connect(parent.openQuoteReplyStatusDialog)
        self.likeStatusAction = QPushButton("Like Status from Accounts...")
        self.likeStatusAction.clicked.connect(parent.openLikeStatusDialog)
        self.boostStatusAction = QPushButton("Boost Status from Accounts...")
        self.boostStatusAction.clicked.connect(parent.openBoostStatusDialog)

        self.statusActionsLayout.addWidget(self.replyStatusAction)
        self.statusActionsLayout.addWidget(self.quoteReplyStatusAction)
        self.statusActionsLayout.addWidget(self.likeStatusAction)
        self.statusActionsLayout.addWidget(self.boostStatusAction)
        self.setLayout(self.statusActionsLayout)


class statusFlagsWidget(QWidget):
    def __init__(self, status, parent=None):
        super(QWidget, self).__init__(parent)

        try:
            if status.inReplyToUri:
                self.replyFlag = QLabel().setPixmap(fp_globals.icons["reply-fill"])
            else:
                self.replyFlag = QLabel()
        except Exception as e:
            print(repr(e))

        self.boostFlag = QLabel()
        # TODO: per-tab boost tracking ffs
        if len(status.boostedBy) > 0:
            self.boostFlag.setPixmap(fp_globals.icons["repeat-line"])

        self.imageFlag = QLabel()
        try:
            if len(status.mpyDict["media_attachments"]):
                if any([i["type"] == "image" for i in status.mpyDict["media_attachments"]]):
                    self.imageFlag.setPixmap(fp_globals.icons["image-fill"])

        except AttributeError:  # status is from DB
            pass

        self.gridLayout = QGridLayout(self)
        self.gridLayout.addWidget(self.replyFlag, 0, 0)
        self.gridLayout.addWidget(self.boostFlag, 0, 1)
        self.gridLayout.addWidget(self.imageFlag, 0, 2)


class statusListDataView(QTreeWidget):
    TIMESTAMP, USER, FLAGS, REPLIES_COUNT, BOOSTS_COUNT, LIKES_COUNT, STATUS_TEXT = range(7)
    def __init__(self):
        QTreeWidget.__init__(self)
        self.newestStatus = None
        self.topLevelStatuses = {}
        self.setColumnCount(7)
        self.setAlternatingRowColors(True)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.openContextMenu)
        self.itemDoubleClicked.connect(self.openStatusDialog)
        self.setColumnWidth(self.TIMESTAMP, 100)
        self.setColumnWidth(self.USER, 175)
        self.setColumnWidth(self.FLAGS, 75)
        self.setColumnWidth(self.REPLIES_COUNT, 2)
        self.setColumnWidth(self.BOOSTS_COUNT, 2)
        self.setColumnWidth(self.LIKES_COUNT, 2)
        self.setColumnWidth(self.STATUS_TEXT, 300)
        self.setHeaderLabels(["Timestamp",
                              "User",
                              "Flags",
                              "R",
                              "B",
                              "L",
                              "Status Text"])

    def statusWidgetFromStatus(self, status, wparent=None):
        #TODO: directly map to defined positions
        columns =  [status.widgetFormattedTimestamp,
                str(status.actor),
                "",
                str(status.replies),
                str(status.boosts),
                str(status.likes),
                str(status.sanitizedContent)]

        if wparent is None:
            wparent = self
        h = QTreeWidgetItem(wparent, columns)
        h.statusUri = status.uri
        h.status = status
        h.setExpanded(True)
        return h

    def doFuckAll(self, *args, **kwargs):
        print("Got callback. Doing fuck all.")
        pass

    def addStatus(self, status):
        if status in self.topLevelStatuses.keys():
            #TODO: Find existing top level widget
            #TODO: self.addMissingChildren(existingWidget)
            return
        h = self.statusWidgetFromStatus(status)
        self.setItemWidget(h, self.FLAGS, statusFlagsWidget(status))
        self.addMissingChildren(h)
        self.sortItems(self.TIMESTAMP, Qt.DescendingOrder)
        self.topLevelStatuses[status] = h
        try:
            self.updateModel()
        except Exception as e:
            fp_globals.systemMessage("addStatus %s while rebuilding model" % repr(e))

    def openStatusDialog(self, item):
        from fe_gui.subwindows import statusDetailsDialog
        h = statusDetailsDialog(self, item.status)

    def addMissingChildren(self, widget):
        widgetStatus = widget.status
        for replyUri in [ro.uri for ro in widgetStatus.knownReplies]:
            if replyUri in fp_globals.loadedStatuses.keys():
                childWidgetStatus = fp_globals.loadedStatuses[replyUri]
                if not any([replyUri == widget.child(i).statusUri for i in range(widget.childCount())]):
                    h = self.statusWidgetFromStatus(childWidgetStatus, widget)
                    self.setItemWidget(h, self.FLAGS, statusFlagsWidget(childWidgetStatus))
            else:
                print("I would add %s as a widget, but it's not in memory!" % replyUri)

    def updateModel(self, *args, **kwargs):
        def _temp_balete_widget(widgetItem):
            # Remove (hopefully) only non-qt reference to widget so it can be gc'd
            del self.topLevelStatuses[widgetItem.status]
            # Remove item from list, destroy it
            (widgetItem.parent() or self.invisibleRootItem()).removeChild(widgetItem)

        try:
            qftext = self.parent().quickFilterTextBox.text()
        except AttributeError:
            qftext = ''

        everyWidgetIterator = QTreeWidgetItemIterator(self)
        # skip topmost object
        everyWidgetIterator += 1
        while everyWidgetIterator.value():
            widget = everyWidgetIterator.value()
            status = widget.status
            if not len(status.knownReplies):
                self.addMissingChildren(widget)

            # Then, see if it should be hidden

            # Then, see if it should be unhidden

            # Then, apply quickfilter
            widget.mutedByQuickFilter = qftext in status.sanitizedContent

            widget.setHidden(not widget.mutedByQuickFilter)
            # Then, see if it should be highlighted, or trigger some other action
            everyWidgetIterator += 1



    def openContextMenu(self, position):
        indexes = self.selectedIndexes()
        try:
            status = fe_core.fe_types.status.getStatusObjectByURI(self.itemFromIndex(indexes[0]).statusUri)
        except:
            print(self.itemFromIndex(indexes[0]).statusUri)
            raise

        menu = QMenu()
        viewStatusDetailsAction = menu.addAction("View Status...")
        viewActorDetailsAction = menu.addAction("View Actor...")
        openInBrowserAction = menu.addAction("Open Status in Browser...")
        menu.addSeparator()
        openActorTabAction = menu.addAction("Open Actor in Anonymous Tab")
        menu.addSeparator()
        likeFromAccountsAction = menu.addAction("Like from Accounts...")
        boostFromAccountsAction = menu.addAction("Boost from Accounts...")
        menu.addSeparator()
        replyAction = menu.addAction("Reply...")
        quoteReplyAction = menu.addAction("Quote Reply...")
        menu.addSeparator()
        mastodonCrawlStatusAction = menu.addAction("Crawl Status")

        action = menu.exec_(self.viewport().mapToGlobal(position))
        if action == viewStatusDetailsAction:
            from fe_gui.subwindows import statusDetailsDialog
            h = statusDetailsDialog(self, status)
        if action == viewActorDetailsAction:
            from fe_gui.subwindows import actorDetailsDialog
            h = actorDetailsDialog(self, status.actor)
        elif action == openInBrowserAction:
            openBrowserTab(status. uri)
        elif action == openActorTabAction:
            from fe_gui.tabs.mastodon_tabs import mastodonActorTab
            tabArgs = {"context": status.actor,
                       "domain": status.actor.domain}
            h = fp_globals.Application.tabs.addTab(mastodonActorTab(fp_globals.Application, **tabArgs), "...")
            fp_globals.Application.tabs.widget(h).updateTabName()
        elif action == likeFromAccountsAction:
            from fe_gui.subwindows import likeFromAccountsDialog
            h = likeFromAccountsDialog(self, status)
        elif action == boostFromAccountsAction:
            from fe_gui.subwindows import boostFromAccountsDialog
            h = boostFromAccountsDialog(self, status)
        elif action == replyAction:
            from fe_gui.subwindows import newStatusDialog
            h = newStatusDialog(self, inReplyTo=status, quote=False)
        elif action == quoteReplyAction:
            from fe_gui.subwindows import newStatusDialog
            h = newStatusDialog(self, inReplyTo=status, quote=True)
        elif action == mastodonCrawlStatusAction:
            from fe_core.mpy_wrapper import mastodonGetContext
            workers.schedulempyJob(mastodonGetContext, {"status": status}, "statusCrawlQueue", self.updateModel, target="anonymous")


class globalSettingsTreeWidget(QTreeWidget):
    def __init__(self, settingsPaneHolder):
        QTreeWidget.__init__(self)
        self.setMinimumSize(150, 150)
        self.setColumnCount(1)
        self.setRootIsDecorated(False)
        self.setHeaderHidden(True)
        self.applicationSettingsTreeWidgetItem = QTreeWidgetItem(self, [fp_globals.productName])
        self.applicationSettingsTreeWidgetItem.settingsPaneWidget = settingsPaneWidget("General Settings",
                                                                                       "General setup and perferences",
                                                                                       fp_globals.Application.config.values())
        settingsPaneHolder.addWidget(self.applicationSettingsTreeWidgetItem.settingsPaneWidget)

        self.instanceSettingsTreeWidgetItem = QTreeWidgetItem(self, ["Instances"])
        self.instanceSettingsTreeWidgetItem.settingsPaneWidget = settingsPaneWidget("Instance Settings",
                                                                                    "Set default options to be used when working with instances",
                                                                                    fe_core.fe_types.instance.InstanceObject.configObjects)
        settingsPaneHolder.addWidget(self.instanceSettingsTreeWidgetItem.settingsPaneWidget)
        for instance in fp_globals.loadedInstances.values():
            # If the config value isn't the default for InstanceObjects
            # The user set something special for this instance, so it's worth showing
            if all([i.configValue == fe_core.fe_types.instance.InstanceObject.configObjectsByName[i.name].default for i in instance.config.values()]):
                break
            #for i in instance.config.values():
            #    if i.configValue != fe_types.InstanceObject.configObjectsByName[i.name].default:
            #        break
            #else:
            #    break  # Don't show this instance, continue outer loop
            instance.instanceSettingsTreeWidgetItem = QTreeWidgetItem(self, [str(instance)])
            instance.instanceSettingsTreeWidgetItem.settingsPaneWidget = settingsPaneWidget(
                "Instance/%s" % str(instance),
                "Set options to be used when working with this instance",
                instance.config.values())
            settingsPaneHolder.addWidget(instance.instanceSettingsTreeWidgetItem.settingsPaneWidget)

        self.accountSettingsTreeWidgetItem = QTreeWidgetItem(self, ["Accounts"])
        self.accountSettingsTreeWidgetItem.settingsPaneWidget = settingsPaneWidget("Account Settings",
                                                                                   "Set global/default options for logged-in accounts",
                                                                                   fe_core.fe_types.account.AccountObject.configObjects)
        settingsPaneHolder.addWidget(self.accountSettingsTreeWidgetItem.settingsPaneWidget)
        for account in fp_globals.loadedAccounts.values():
            account.accountSettingsTreeWidgetItem = QTreeWidgetItem(self, [str(account)])
            account.accountSettingsTreeWidgetItem.settingsPaneWidget("Account/%s" % str(account),
                                                                     "Set options to be used for this account",
                                                                     account.config.values())
            settingsPaneHolder.addWidget(account.accountSettingsTreeWidgetItem.settingsPaneWidget)

        self.setCurrentItem(self.applicationSettingsTreeWidgetItem)


@attr.s
class settingsPaneWidget(QWidget):
    title = attr.ib()
    description = attr.ib()
    context = attr.ib()
    showSaveDiscardButtons = attr.ib(default=False)

    def __attrs_post_init__(self):
        QWidget.__init__(self)
        self.configWidgets = {}
        self.setVisible(False)
        self.settingsPaneLayout = QVBoxLayout()
        self.settingsPaneLayout.addWidget(titleAndDescriptionLabelWidget(self))
        for i in self.context:
            self.settingsPaneLayout.addWidget(configObjectWidget(self, i))
        self.settingsPaneLayout.addStretch()
        if self.showSaveDiscardButtons:
            self.saveDiscardButtons = QDialogButtonBox(QDialogButtonBox.Apply | QDialogButtonBox.Discard | QDialogButtonBox.RestoreDefaults)
            self.saveDiscardButtons.button(QDialogButtonBox.Apply).clicked.connect(self.applyChanges)
            self.saveDiscardButtons.button(QDialogButtonBox.Discard).clicked.connect(self.discardChanges)
            self.saveDiscardButtons.button(QDialogButtonBox.RestoreDefaults).clicked.connect(self.clearObjectConfig)
            self.settingsPaneLayout.addWidget(self.saveDiscardButtons)
        self.setLayout(self.settingsPaneLayout)

    def changedWidgets(self):
        for cWidget in self.configWidgets.values():
            try:
                if cWidget.control != cWidget.configObject.configValue:
                    print("%s has been changed to %s from its original %s" % (cWidget.configObject.name, cWidget.control, cWidget.configObject.configValue))
                    yield cWidget
            except AttributeError: #Editing type defaults
                if cWidget.control != cWidget.configObject.default:
                    print("%s has been changed to %s from its original %s" % (cWidget.configObject.name, cWidget.control, cWidget.configObject.default))
                    yield cWidget

    def clearObjectConfig(self):
        for c in self.context:
            c.resetToUserDefault()

    def discardChanges(self):
        for cWidget in self.changedWidgets():
            try:
                cWidget.control = cWidget.configObject.configValue
            except AttributeError:
                cWidget.control = cWidget.configObject.default

    def applyChanges(self):
        for cWidget in self.changedWidgets():
            print("trying to save %s/%s" % (cWidget.configObject.context, cWidget.configObject.name))
            try:
                h = cWidget.configObject.configValue
                cWidget.configObject.dirty = True
                cWidget.configObject.configValue = cWidget.control
            except AttributeError: #Editing type defaults
                cWidget.configObject.default = cWidget.control
                fp_globals.dirtyGlobalConfigObjects.add(cWidget.configObject)
            print("settings pushed to object")

    def anyWidgetStateChanged(self):
        pass


@attr.s
class titleAndDescriptionLabelWidget(QWidget):
    context = attr.ib()

    def __attrs_post_init__(self):
        QWidget.__init__(self)
        layout = QVBoxLayout()

        titleLabelWidget = QLabel(self.context.title)
        titleLabelWidget.setStyleSheet("font: 16pt")
        # titleLabelWidget.setMaximumHeight(20)
        descriptionLabelWidget = QLabel(self.context.description)
        # descriptionLabelWidget.setMaximumHeight(15)
        layout.addWidget(titleLabelWidget)
        layout.addWidget(descriptionLabelWidget)
        self.setLayout(layout)


@attr.s
class configObjectWidget(QWidget):
    parent = attr.ib()
    configObject = attr.ib()

    def __attrs_post_init__(self):
        QWidget.__init__(self)
        self.dirty = False
        self.layout = QHBoxLayout()
        if type(self.configObject) == type:
            # Setting "defaults" means changing the class, so it's not instantiated yet
            self.control = self.configObject.default
        else:
            self.control = self.configObject.configValue

        self.layout.addWidget(QLabel(self.configObject.displayString))
        self.layout.addStretch()

        if self.configObject.selectionType == "bool":
            self.configBoolCheckbox = QCheckBox()
            self.configBoolCheckbox.setChecked(self.control)
            self.configBoolCheckbox.stateChanged.connect(self.thisWidgetStateChanged)
            self.configBoolCheckbox.stateChanged.connect(self.parent.anyWidgetStateChanged)
            self.layout.addWidget(self.configBoolCheckbox)
        elif self.configObject.selectionType == "int":
            self.configIntSpinBox = QSpinBox()
            self.configIntSpinBox.setValue(self.control)
            self.configIntSpinBox.valueChanged.connect(self.thisWidgetStateChanged)
            self.configIntSpinBox.valueChanged.connect(self.parent.anyWidgetStateChanged)
            self.layout.addWidget(self.configIntSpinBox)
        elif self.configObject.selectionType == "str":
            self.configLineEdit = QLineEdit()
            self.configLineEdit.setText(self.control)
            self.configLineEdit.textChanged.connect(self.thisWidgetStateChanged)
            self.configLineEdit.textChanged.connect(self.parent.anyWidgetStateChanged)
            self.layout.addWidget(self.configLineEdit)
        elif self.configObject.selectionType == "dynamicChoice":
            self.configComboBox = QComboBox()
            for i in self.configObject.dynamicChoiceList.values():
                self.configComboBox.addItem(str(i), i)
            #self.configComboBox.setText(self.control)
            self.configComboBox.currentIndexChanged.connect(self.thisWidgetStateChanged)
            self.configComboBox.currentIndexChanged.connect(self.parent.anyWidgetStateChanged)
        else:
            self.placeholderLabel = QLabel("?")
            self.layout.addWidget(self.placeholderLabel)

        self.setLayout(self.layout)
        self.parent.configWidgets[self.configObject.name] = self

    def thisWidgetStateChanged(self, state):
        self.dirty = True
        selType = self.configObject.selectionType
        if selType == "bool":
            self.control = bool(state)
        elif selType == "int":
            self.control = int(state)
        elif selType == "str":
            self.control = str(state)
        elif selType == "dynamicChoice":
            #TODO: Ensure we're fetching object from UserRole
            #https://doc.qt.io/qt-5/qcombobox.html#editTextChanged
            self.control = self.configComboBox.itemData(state)


class quickPostLineEdit(QLineEdit):
    def __init__(self, parent, status):
        QLineEdit.__init__(self, parent)
        #TODO: Add flag for mention (on actor details, for example)/other uses, or subclass
        #TODO: _handle_of_default_posting_account
        #TODO: wire returnpressed function
        self.setPlaceholderText("Quick reply from _handle_of_default_posting_account")
        self.returnPressed.connect(self.postReply)

    def postReply(self):
        pass


class actorSmallIconWidget(QLabel):
    def __init__(self, parent, actor):
        QLabel.__init__(self, parent)
        self.iPixmap = QPixmap()
        try:
            self.iPixmap.loadFromData(actor.meta[fe_core.fe_types.actor.ActorMetadataType.ICON_STATIC_BLOB].data)
            self.setPixmap(self.iPixmap)
        except KeyError:
            #TODO: Placeholder avatar
            pass


class actorNameplateWidget(QWidget):
    def __init__(self, parent, actor, icon=False):
        QWidget.__init__(self, parent)
        self.l1layout = QVBoxLayout()
        #TODO: Sanitized display name
        self.l1actorDisplayNameLabel = QLabel(actor.displayName)
        self.l1actorDisplayNameLabel.setStyleSheet("font: 16pt")
        self.l1actorHandleLabel = QLabel(str(actor))
        self.l1layout.addWidget(self.l1actorDisplayNameLabel)
        self.l1layout.addWidget(self.l1actorHandleLabel)
        if icon:
            self.layout = QHBoxLayout()
            self.layout.addWidget(actorSmallIconWidget(self, actor))
            self.layout.addLayout(self.l1layout)
            self.layout.addStretch()
            self.setLayout(self.layout)
        else:
            self.setLayout(self.l1layout)


class actorDetailsWidget(QWidget):
    def __init__(self, parent, actor):
        QWidget.__init__(self, parent)

        self.actor = actor
        self.layout = QVBoxLayout()
        self.actorNameplateWidget = actorNameplateWidget(self, actor)

        self.actorSummaryTextBox = QPlainTextEdit()
        self.actorSummaryTextBox.setReadOnly(True)
        self.actorSummaryTextBox.setMinimumWidth(500)
        self.actorSummaryTextBox.appendHtml(actor.summary)
        # form fields (on sidebar?)

        # timeline view with known statuses
        self.actorTimelineDataView = statusListDataView()

        self.layout.addWidget(self.actorNameplateWidget)
        self.layout.addWidget(self.actorSummaryTextBox)
        self.layout.addWidget(self.actorTimelineDataView)
        self.setLayout(self.layout)


class statusDetailsWidget(QWidget):
    def __init__(self, parent, status):
        QWidget.__init__(self, parent)
        self.status = status

        if fp_globals.loadedAccounts:
            self.quickPostLineEdit = quickPostLineEdit(self, self.status)
        else:
            self.quickPostLineEdit = QLineEdit("Add an account to use Quick Replies")
            self.quickPostLineEdit.setEnabled(False)

        self.actorNameplateWidget = actorNameplateWidget(self, status.actor, icon=True)
        # Layer 2, timestamp and numbers
        self.l2layout = QHBoxLayout()
        self.l2timestamp = QLabel()
        self.l2timestamp.setStyleSheet("text-decoration: underline")
        self.l2timestamp.setText(str(status.timestamp))
        self.l2timestamp.setToolTip(status.uri)
        self.l2layout.addWidget(self.l2timestamp)

        # Layer 3, Content
        self.l3 = QPlainTextEdit()
        self.l3.setReadOnly(True)
        self.l3.setMinimumWidth(500)
        self.l3.appendHtml(str(status))

        # Layer 4, images
        self.l3layout = QHBoxLayout()
        if len(status.images):
            imageWidgets = []
            for i in status.images:
                iw = cachedImageThumbnailWidget(self, i)
                create_task(iw.load())
                imageWidgets.append(iw)

            for i in imageWidgets:
                self.l3layout.addWidget(i)

        self.layout = QVBoxLayout(self)
        self.layout.addWidget(self.actorNameplateWidget)
        self.layout.addLayout(self.l2layout)
        self.layout.addWidget(self.l3)
        self.layout.addWidget(self.quickPostLineEdit)
        self.layout.addLayout(self.l3layout)
        #self.setLayout(self.layout)


class objectDetailsTabWidget(QTabWidget):
    def __init__(self, parent, context):
        QTabWidget.__init__(self, parent)

        self.addTab(parent.objectDetailsWidget, "%s" % str(type(context)))
        self.addTab(settingsPaneWidget("%s Settings\%s" % (str(type(context)), context.handle),
                                       "Override global defaults for this %s" % str(type(context)),
                                       context), "Settings")


class accountSelector(QListView):
    def __init__(self, parent, accounts=None):
        QListView.__init__(self)

        model = QStandardItemModel(self)

        self.accountItems = {}
        for i, accObj in fp_globals.loadedAccounts.items():
            h = QStandardItem(i)
            h.setCheckable(True)
            if accounts is None:
                if accObj.config["preselectInAccountSelector"].configValue:
                    h.setCheckState(True)
            else:
                if i in accounts or accObj in accounts:
                    h.setCheckState(True)
            self.accountItems[i] = h
            model.appendRow(h)

        self.setModel(model)


class visibilityScopeSelectorComboBox(QComboBox):
    def __init__(self, parent):
        QComboBox.__init__(self, parent)
        for i in fe_core.fe_types.common.VisibilityScope:
            self.addItem(i.name)


class accountSelectorComboBox(QComboBox):
    def __init__(self, parent, noanon=False):
        QComboBox.__init__(self, parent)
        model = QStandardItemModel(self)
        if not noanon:
            model.appendRow(QStandardItem("(Anonymous)"))
        for i in fp_globals.loadedAccounts.keys():
            h = QStandardItem(i)
            model.appendRow(h)

        self.setModel(model)

    def selectedAccount(self):
        return self.model().itemData(self.currentIndex())
